"""
This module handles the actual assembly of the run report. It converts each of
the step reports into a ``StepReport`` object and then converts the run
to a ``Report`` object containing the list of ``StepReport`` objects.

It returns this ``Report`` object as the final product of the run.
"""

from __future__ import annotations

import datetime
import logging
import time
from collections import OrderedDict, defaultdict
from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Optional, Tuple, Union

from thoughtful.supervisor.manifest import StepId
from thoughtful.supervisor.reporting.record import Record
from thoughtful.supervisor.reporting.record_report import RecordReport
from thoughtful.supervisor.reporting.report import Report
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.step_report import StepReport
from thoughtful.supervisor.reporting.timer import Timer

logger = logging.getLogger(__name__)


@dataclass
class RecordAccumulator:
    """
    Manages a list of records that are being processed by a step.
    """

    records_by_id: Dict[RecordId, Record] = field(default_factory=OrderedDict)

    def upsert(self, record: Record):
        """Insert a new Record, or update the record if it already exists."""
        self.records_by_id[record.record_id] = record

    def exists(self, _id: RecordId) -> bool:
        return _id in self.records_by_id.keys()

    def soft_update(self, record: Record):
        """
        Update a record *only* if it doesn't exist *or* the current status is
        RUNNING, otherwise do nothing.
        """
        if (
            self.exists(record.record_id)
            and self.records_by_id[record.record_id].status != Status.RUNNING
        ):
            return
        self.upsert(record)

    def to_reports(self, base_step_report: StepReport) -> List[RecordReport]:
        return [
            RecordReport.from_step_report(base_step_report, record)
            for record in self.records_by_id.values()
        ]

    def __iter__(self) -> Iterable[Record]:
        return iter(self.records_by_id.values())


@dataclass
class StepReportBuilder:
    """
    Builds a step report during the lifecycle of a step's execution. Once a step
    has finished executing, you can call ``_to_report`` to get the final
    ``StepReport``.
    """

    step_id: str
    """
    str: The ID of the step.
    """

    start_time: datetime.datetime
    """
    datetime.datetime: The start time of the step.
    """

    status: Status
    """
    Status: The status of the step.
    """

    end_time: Optional[datetime.datetime] = None
    """
    datetime.datetime: The end time of the step.
    """

    _record_accumulator: RecordAccumulator = field(default_factory=RecordAccumulator)
    """
    RecordAccumulator: Holds records that were processed by this step.
    """

    def update_from(self, other: StepReportBuilder) -> None:
        """
        Update a new ``StepReportBuilder`` from an existing ``StepReportBuilder``.

        Args:
            other (StepReportBuilder): The step report builder to update from.
        """
        # this bypasses StepReports created by decorator method (ie. do not have
        # records)
        other_records_by_id = {r.record_id: r for r in other.records}
        for record in self.records:
            if record.record_id in other_records_by_id:
                logger.warning(
                    f"Duplicate record '{record.record_id}' found in step '{self.step_id}'"
                )

        self.end_time = other.end_time
        self.status = other.status
        for other_record in other.records:
            self._record_accumulator.upsert(other_record)

    def to_reports(self) -> List[Union[StepReport, RecordReport]]:
        step_report = self._to_report()
        record_reports: List[
            Union[StepReport, RecordReport]
        ] = self._record_accumulator.to_reports(step_report)
        return record_reports + [step_report]

    def set_record_status(
        self,
        record_id: RecordId,
        status: Status,
        message: Optional[str] = None,
        metadata: Optional[dict] = None,
        is_soft_update: bool = False,
    ):
        message = message or ""
        metadata = metadata or {}
        new_record = Record(record_id, status, message, metadata)
        if is_soft_update:
            self._record_accumulator.soft_update(new_record)
        else:
            self._record_accumulator.upsert(new_record)

    @property
    def records(self) -> Tuple[Record]:
        return tuple(self._record_accumulator)

    def _to_report(self) -> StepReport:
        """
        An easily "jsonable" final report on this step's execution.

        Returns:
            StepReport: A final report on this step's execution.
        """

        # Build the report
        return StepReport(
            step_id=self.step_id,
            status=self.status,
            start_time=self.start_time,
            end_time=self.end_time,
        )


RecordId = str
StepId = str


@dataclass
class ReportBuilder:
    """
    A work report builder that creates a new work report as a digital worker
    is executed.
    """

    timer: Timer = field(default_factory=Timer)
    """
    Timer: The timer used to time the execution of the workflow.
    """

    _step_report_builders: List[StepReportBuilder] = field(default_factory=list)
    """
    List[StepReportBuilder]: The list of step reports.
    """

    timer_start: float = time.perf_counter()
    """
    float: The start time of the workflow.
    """

    status: Optional[Status] = None
    """
    Status, optional: The status of the run.
    """

    # These steps will be overridden with the specified status when the
    # `Report` is written
    _step_statuses_to_override: Dict[StepId, Status] = field(default_factory=dict)
    """
    Dict[StepId, Status]: The statuses to override for each step
    """
    _records_to_override: Dict[StepId, Dict[RecordId, Record]] = field(
        default_factory=lambda: defaultdict(dict)
    )
    """
    Dict[StepId, Dict[RecordId, Status]]: The statuses to override for each
    record
    """

    run_had_exception: bool = False
    """
    Boolean value to indicate if the run terminated on an exception.
    """

    _run_status_override: Optional[Status] = None
    """
    Override the status of the run to be in the status of `status`.
    """

    def __post_init__(self):
        self.timer.start()

    def fail_step(self, step_id: str) -> None:
        """
        Override a step to be in the `StepStatus.ERROR` state. Note: this
        overrides every step with this ID, so if the step ran multiple times
        in the workflow, they will all be marked as failed.

        Args:
            step_id (str): The step id to override.
        """
        self.set_step_status(step_id=step_id, status=Status.FAILED)

    def set_step_status(self, step_id: str, status: Union[Status, str]) -> None:
        """
        Override a step to be in the status of `status`. Note: this
        overrides every step with this ID, so if the step ran multiple times
        in the workflow, they will all be marked as this `status`.

        Args:
            step_id (str): The step id to override.
            status (Status | str): The status to override the step to.
        """
        # Convert the status to the correct type if necessary
        safe_status = Status(status)
        self._step_statuses_to_override[step_id] = safe_status

    def find_index(self, step_report_builder: StepReportBuilder) -> Union[int, None]:
        """
        Find the index of the step_report_builder in the ReportBuilder's _step_report_builders
        that matches the incoming steps and does not already have Records attached to it.
        """
        index = next(
            (
                index
                for index, step in enumerate(self._step_report_builders)
                if step.step_id == step_report_builder.step_id
            ),
            None,
        )
        return index

    def add_step_report(self, step_report_builder: StepReportBuilder) -> None:
        """
        Add a StepReport if it doesn't exist already. If it does exist, update
        the existing StepReport with the new StepReport's records.
        """
        index = self.find_index(step_report_builder)
        if index is not None or index == 0:
            self._step_report_builders[index].update_from(step_report_builder)
        else:
            self._step_report_builders.append(step_report_builder)

    def set_record_status(
        self,
        step_id: str,
        record_id: str,
        status: Union[Status, str],
        message: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> None:
        """
        Override a record to be in the status of `status`. Note: this
        overrides every step with this step ID and this record ID, so if the
        step ran multiple times in the workflow, they will all be marked as
        this `status`.

        Args:
            step_id (str): The step id a specific step that contains this record
            record_id (str): The id of the record to override.
            status (Status | str): The status to override the record to.
        """
        # Convert the status to the correct type if necessary
        if type(step_id) != str:
            raise TypeError("step_id must be a string")
        if type(record_id) != str:
            raise TypeError("record_id must be a string")
        message = message or ""
        metadata = metadata or {}
        safe_status = Status(status)
        self._records_to_override[step_id][record_id] = Record(
            record_id=record_id, status=safe_status, message=message, metadata=metadata
        )

    def set_run_status(self, status: Union[Status, str]) -> None:
        """
        Manually set the status of the bot run. If not set, the run
        status will be determined automatically

        Args:
            status (Union[Status, str]): The status to override the run to.
        """
        # Convert the status to the correct type if necessary
        safe_status = Status(status)
        self._run_status_override = safe_status

    def to_report(self) -> Report:
        """
        Convert supervisor workflow to work report. It is here that we
        convert the entire workflow to a list of ``StepReport`` objects.
        After which, we pass over the entire list overriding the record
        and step statuses according to the ``_step_statuses_to_override``
        and ``_records_to_override`` dictionaries.

        Returns:
            Report: The finalized work report.
        """
        timed = self.timer.end()

        for step_builder in self._step_report_builders:
            # Override the step status if requested
            if step_builder.step_id in self._step_statuses_to_override:
                new_status = self._step_statuses_to_override[step_builder.step_id]
                step_builder.status = new_status
            # Override any of the step's record statuses if requested
            if step_builder.step_id in self._records_to_override:
                records = self._records_to_override[step_builder.step_id]
                # NOTE: need to make change here from status - ReportBuilder saving
                # Records instead of just status - in self._records_to_override
                for record_id, record in records.items():
                    step_builder.set_record_status(
                        record_id, record.status, record.message, record.metadata
                    )

        # Merge all the step reports together
        final_workflow = [
            report
            for step in self._step_report_builders
            for report in step.to_reports()
        ]

        # Set the run status
        self.status = Status.FAILED if self.run_had_exception else Status.SUCCEEDED
        if self._run_status_override is not None:
            self.status = self._run_status_override

        return Report(
            start_time=timed.start,
            end_time=timed.end,
            workflow=final_workflow,
            status=self.status,
        )



"""
This module contains the ``Record`` class, a non-user facing class that is used
to store information about a step's record. This class is represented in the
run report as a json object.

.. code-block:: json

    {
        "workflow": [
            {
                "step_id": "1.1",
                "step_status": "succeeded",
                "record": {
                    "id": "1",
                    "status": "succeeded",
                    "message": "Message can provide additional information",
                    "metadata": {
                        "internal_reason": "record amount is too low",
                        "amount": 15
                    }
                }
            }
        ]
    }

"""
import json
import logging
import warnings
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from thoughtful.supervisor.reporting.status import Status


@dataclass
class Record:
    """
    A record can refer to any object that is being processed by a step. If a
    step in your workflow is repeating many times for different items, then
    each item you're iterating over can be a record.
    """

    record_id: str
    """
    str: The status of the record.
    """

    status: Status
    """
    Status: The status of the record.
    """

    message: Optional[str] = field(default_factory=str)
    """
    Message: A message that can provide additional information about the record,
    such as a (short) description of why the record is in a certain state.
    """

    # make required default = {} ?
    metadata: Optional[dict] = field(default_factory=dict)
    """
    Metadata: A dictionary of additional information about the record. This is
    mostly meant to "tag" the record for later analysis and grouping.
    """

    def __post_init__(self):
        """
        Validate the record message length
        """
        if len(self.message) > 120:
            raise ValueError("Record Message Field - 120 Character limit exceeded")
        validate_metadata_json_helper(self.metadata)

    def __json__(self):
        return {
            "id": self.record_id,
            "status": self.status.value,
            "message": self.message,
            "metadata": self.metadata,
        }


def validate_metadata_json_helper(item: Dict):
    """
    Validates that the metadata field is a valid JSON object and that it is less
    than 5 KB
    """
    try:
        json_str = json.dumps(item)
        if len(json_str) / 1024 > 5:
            warnings.simplefilter("always", UserWarning)
            warnings.warn(
                "Record metadata validation failed - Object should not exceed 5 KB",
                category=UserWarning,
            )
            warnings.resetwarnings()
    except Exception as e:
        logging.exception(e)
        warnings.simplefilter("always", UserWarning)
        warnings.warn(
            "Could not serialize record metadata into JSON", category=UserWarning
        )
        warnings.resetwarnings()



from dataclasses import dataclass, field
from typing import Callable, List, Union

from thoughtful.supervisor.manifest import Manifest
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.step_report import StepReport


@dataclass
class StepReportChangeEvent:
    step_report: StepReport


@dataclass
class NewManifestEvent:
    manifest: Manifest


@dataclass
class RunStatusChangeEvent:
    status: Status


@dataclass
class ArtifactsUploadedEvent:
    output_uri: str


Event = Union[
    StepReportChangeEvent,
    NewManifestEvent,
    RunStatusChangeEvent,
    ArtifactsUploadedEvent,
]


@dataclass
class EventBus:
    subscribers: List[Callable[[Event], None]] = field(default_factory=list)

    def subscribe(self, callback: Callable[[Event], None]):
        self.subscribers.append(callback)

    def emit(self, event: Event):
        for subscriber in self.subscribers:
            subscriber(event)



"""
This module contains the code for the MainContext class. This class is
a context manager that supervises the execution of the main body of the process.

On entrance, it will verify the process manifest.

Upon exit, it will convert the step report into the run report. It will also
evaluate the ultimate status of the process—success if no exceptions were
raised, failure otherwise.

Without this context, the output manifest and run report would not be
generated at the end of the process.
"""

from __future__ import annotations

import datetime
import logging
import os
import pathlib
from types import TracebackType
from typing import Callable, Optional, Type, Union
from urllib.parse import urlparse

import boto3

from thoughtful.supervisor.event_bus import ArtifactsUploadedEvent, EventBus
from thoughtful.supervisor.event_bus import NewManifestEvent, RunStatusChangeEvent
from thoughtful.supervisor.manifest import Manifest
from thoughtful.supervisor.reporting.report_builder import ReportBuilder
from thoughtful.supervisor.reporting.status import Status

logger = logging.getLogger(__name__)


class MainContext:
    """
    Supervises an entire digital worker run and generates a work report
    and drift report for the run from the digital worker's manifest.

    You can optionally specify a callback function that will be run when
    this context is finished via the `callback` param in the constructor. A
    callback is a function that is invoked with two parameters: the
    current context (as the `MainContext` instance) and the `Report`
    generated from this digital worker's run.

    For example:

    .. code-block:: python

        def print_work_report(
            ctx: MainContext,
            work_report: Report
        ):
            print(work_report.__json__())

        def main()
            # ...

        with supervise(callback=print_work_report):
            main()
    """

    def __init__(
        self,
        report_builder: ReportBuilder,
        manifest: Union[Manifest, str, pathlib.Path],
        output_dir: Union[str, pathlib.Path],
        event_bus: EventBus,
        upload_uri: Optional[str] = None,
        callback: Optional[Callable] = None,
    ):
        """
        Args:
            report_builder (ReportBuilder): A ReportBuilder object that will
                receive the step reports and provide the run report.
                messages and logs throughout the process execution.
            manifest (str, Path): A pathlike object that points to the manifest
                file for the process.
            output_dir (str, Path): A pathlike object that points to the output
                directory for the process. This will receive the run report and
                output manifest.
            event_bus (EventBus): This instance will send events, such as
                a new manifest or a change in the run's status, to this bus.
            upload_uri (str, optional): A URI to upload the output files to.
            callback (callable, optional): a function that is invoked with three
                parameters: the current context (as the `MainContext` instance)
                and the `Report` generated from this digital worker's run.
        """
        self.report_builder = report_builder
        self.output_path = pathlib.Path(output_dir)
        self.upload_uri = upload_uri

        self.manifest_path = (
            manifest if isinstance(manifest, (str, pathlib.Path)) else None
        )
        self.manifest = self._parse_manifest(manifest)
        self.callback = callback
        self.event_bus = event_bus

    @staticmethod
    def _parse_manifest(
        manifest: Union[Manifest, str, pathlib.Path]
    ) -> Optional[Manifest]:
        if isinstance(manifest, Manifest):
            return manifest
        manifest_path = pathlib.Path(manifest)
        try:
            manifest = Manifest.from_file(manifest_path)
            return manifest
        except Exception:
            logger.exception("warning: could not read manifest")
        return None

    def __enter__(self) -> MainContext:
        """
        Logic for when this context is first started. Attempts to load the
        manifest and returns itself as the context.

        Returns:
            MainContext: This instance.
        """
        if self.manifest:
            self.event_bus.emit(
                NewManifestEvent(
                    manifest=self.manifest,
                )
            )
        self._stream_run_status_change(Status.RUNNING)

        return self

    def set_run_status(self, status: Union[Status, str]) -> None:
        """
        Sets the run status of the process. This is the final status of the
        process, and will be used to determine the status of the run report.

        Args:
            status (Union[Status, str]): The status to set.
        """
        self.report_builder.set_run_status(status)
        self._stream_run_status_change(status)

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """
        Runs when the context is about to close, whether caused
        by a raised Exception or now.

        Returns:
            bool: True if the parent caller should ignore the
                Exception raised before entering this function
                (if any), False otherwise.
        """
        if exc_type:
            self.report_builder.run_had_exception = True
        work_report = self.report_builder.to_report()

        self._stream_run_status_change(work_report.status)

        # Create the output directory if it doesn't already exist
        self.output_path.mkdir(parents=True, exist_ok=True)

        # Write the work report
        work_report_path = self._safe_report_path(file_prefix="run-report")
        work_report.write(work_report_path)

        # Write the manifest back out as a JSON file
        if self.manifest:
            manifest_json_path = self.output_path / "manifest.json"
            self.manifest.write_to_json_file(manifest_json_path)

        # Run the user-defined callback
        if self.callback:
            self.callback(self, work_report)

        # Upload output files to S3
        if self.upload_uri:
            try:
                self._upload_output_files_to_s3(self.upload_uri)
            except Exception:
                logger.exception("Failed to upload output files to S3")
        elif os.environ.get("ROBOCORP_HOME") is None:
            logger.warning(
                "SUPERVISOR_ARTIFACT_UPLOAD_URI is not set. Artifacts"
                " will not be uploaded to S3."
            )

        # Explicitly return false so that the parent caller above
        # this context sees the exception (if any)
        return False

    def _stream_run_status_change(self, status: Status) -> None:
        """
        Post a status change to the stream callback if it exists.
        """
        self.event_bus.emit(RunStatusChangeEvent(status=status))

    def _upload_output_files_to_s3(self, upload_uri: str) -> None:
        """
        It uploads all files in the output directory to S3. It requires the
        environment variable `SUPERVISOR_ARTIFACT_UPLOAD_URI` to be set with
        the S3 URI to upload the files to.

        Args:
            upload_uri (str): The S3 URI to upload the files to.
        """
        s3_client = boto3.client("s3")
        parsed_upload_uri = urlparse(upload_uri.strip())
        bucket = parsed_upload_uri.hostname
        path = parsed_upload_uri.path.strip("/")

        for file in self.output_path.glob("*"):
            try:
                if file.is_file():
                    obj = f"{path}/{file.name}" if path else file.name
                    s3_client.upload_file(str(file), bucket, obj)
            except Exception:
                logger.exception(f"Failed to upload {file} to S3")
        self.event_bus.emit(ArtifactsUploadedEvent(output_uri=upload_uri))

    def _safe_report_path(self, file_prefix: str) -> pathlib.Path:
        """
        A ``pathlib.Path`` instance that points to a new work report writable
        location that is safe across all OSes.

        Returns:
            pathlib.Path: The path to the new report to be written.
        """
        timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S")
        filename = f"{file_prefix}-{timestamp}.json"

        # Remove any characters from the timestamp that OSes don't like
        invalid_chars = [":", "*", "?", '"', "<", ">" "|", "'"]
        for char in invalid_chars:
            filename = filename.replace(char, "_")

        return self.output_path / filename



"""
Defines json payloads to be delivered to the streaming endpoint
"""
import datetime
from dataclasses import dataclass

import thoughtful
from thoughtful.supervisor.manifest import Manifest
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.step_report import StepReport
from thoughtful.supervisor.streaming.action import Action


@dataclass
class Payload:
    """
    Base payload format for all streaming messages.
    """

    run_id: str
    action: Action
    client: str = "supervisor"
    version: str = thoughtful.__version__

    def __json__(self) -> dict:
        return {
            "run_id": self.run_id,
            "client": self.client,
            "version": self.version,
            "action": self.action.value,
            "payload": {
                "timestamp": datetime.datetime.now(datetime.timezone.utc).isoformat()
            },
        }


class StepReportPayload(Payload):
    """
    Payload for sending a step report action.
    """

    def __init__(self, step_report: StepReport, run_id: str):
        super().__init__(run_id=run_id, action=Action.STEP_REPORT)
        self.step_report = step_report

    def __json__(self) -> dict:
        _json = super().__json__()
        _json["payload"]["step_report"] = self.step_report.__json__()
        return _json


class BotManifestPayload(Payload):
    """
    Payload for sending the manifest.
    """

    def __init__(self, manifest: Manifest, run_id: str):
        self.manifest = manifest
        super().__init__(run_id=run_id, action=Action.BOT_MANIFEST)

    def __json__(self) -> dict:
        _json = super().__json__()
        _json["payload"]["bot_manifest"] = self.manifest.__json__()
        return _json


class ArtifactsUploadedPayload(Payload):
    """
    Payload for notifying the stream consumer that artifacts have been uploaded
    to S3.
    """

    def __init__(self, run_id: str, output_artifacts_uri: str):
        self.output_artifacts_uri = output_artifacts_uri
        super().__init__(run_id, action=Action.ARTIFACTS_UPLOADED)

    def __json__(self) -> dict:
        _json = super().__json__()
        _json["payload"]["output_artifacts_uri"] = self.output_artifacts_uri
        return _json


class RunStatusChangePayload(Payload):
    """
    Payload for notifying the stream consumer that a bot's status has changed
    """

    def __init__(self, run_id: str, status: Status):
        self.status = status
        super().__init__(run_id, action=Action.STATUS_CHANGE)

    def __json__(self) -> dict:
        _json = super().__json__()
        _json["payload"]["status"] = self.to_fabric_process_status(self.status)
        return _json

    @staticmethod
    def to_fabric_process_status(status: Status) -> str:
        if status == Status.RUNNING:
            return "processing"
        elif status == Status.FAILED:
            return "failed"
        elif status == Status.SUCCEEDED:
            return "finished"
        elif status == Status.WARNING:
            return "warning"
        else:
            raise ValueError(f"Unknown status: {status}")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict

from thoughtful.supervisor.reporting.record import Record
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.step_report import StepReport
from thoughtful.supervisor.reporting.timed_report import TimedReport


@dataclass
class RecordReport(TimedReport):
    step_id: str
    """
    str: The ID of the step.
    """

    status: Status
    """
    Status: The status of the step.
    """

    record: Record
    """
    Record: The record that was processed by the step.
    """

    @classmethod
    def from_step_report(cls, step_report: StepReport, record: Record) -> RecordReport:
        return cls(
            start_time=step_report.start_time,
            end_time=step_report.end_time,
            step_id=step_report.step_id,
            status=step_report.status,
            record=record,
        )

    def __json__(self) -> Dict[str, Any]:
        return {
            **super().__json__(),
            "step_id": self.step_id,
            "step_status": self.status.value,
            "record": self.record.__json__(),
        }



from __future__ import annotations

import datetime

import jwt


class Token:
    """A JWT token."""

    def __init__(self, encoded_value: str, algorithm="HS256"):
        self.encoded_value = encoded_value

        decoded_values = jwt.decode(
            jwt=self.encoded_value,
            algorithms=[algorithm],
            options={"verify_signature": False},
        )
        self.expires_at = datetime.datetime.fromtimestamp(decoded_values["exp"])
        self.issued_at = datetime.datetime.fromtimestamp(decoded_values["iat"])

    def __str__(self):
        return self.encoded_value

    @property
    def is_expired(self, expiry_margin: float = 0.1) -> bool:
        """
        Returns True if the token is expired or will expire soon.

        expiry_margin: A float between 0 and 1. For example, if set to 0.1, the
            token will be considered expired if it will expire within 10% of its
            total lifetime.
        """
        now = datetime.datetime.now()
        if now >= self.expires_at:
            return True

        time_to_expire_seconds = (self.expires_at - self.issued_at).total_seconds()
        expiry_threshold = time_to_expire_seconds * expiry_margin
        will_expire_soon = abs(time_to_expire_seconds) <= expiry_threshold
        if will_expire_soon:
            return True

        return False



from __future__ import annotations

import logging
from typing import Optional

import requests

from thoughtful.supervisor.event_bus import ArtifactsUploadedEvent, Event
from thoughtful.supervisor.event_bus import NewManifestEvent, RunStatusChangeEvent
from thoughtful.supervisor.event_bus import StepReportChangeEvent
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.streaming.jwt_auth import JWTAuth
from thoughtful.supervisor.streaming.payloads import ArtifactsUploadedPayload
from thoughtful.supervisor.streaming.payloads import BotManifestPayload, Payload
from thoughtful.supervisor.streaming.payloads import RunStatusChangePayload
from thoughtful.supervisor.streaming.payloads import StepReportPayload
from thoughtful.supervisor.streaming.token import Token

logger = logging.getLogger(__name__)

POST_TIMEOUT_SECONDS = 10


class Streamer(requests.Session):
    """
    Provides run status updates to a third party client (eg, Fabric/Empower)
    by posting requests to its callback URL.
    """

    def __init__(
        self,
        run_id: str,
        callback_url: str,
        auth: JWTAuth,
    ):
        super().__init__()
        self.run_id = run_id
        self.callback_url = callback_url
        self.auth: JWTAuth = auth

    @classmethod
    def from_encoded_tokens(
        cls,
        run_id: str,
        callback_url: str,
        access_token: str,
        refresh_token: str,
        refresh_url: str,
    ) -> Streamer:
        """
        Convenience constructor for creating an instance from the string
        (encoded) JWT tokens.
        """
        new_auth = JWTAuth(
            access_token=Token(access_token) if access_token else None,
            refresh_token=Token(refresh_token) if refresh_token else None,
            refresh_url=refresh_url,
        )
        return cls(
            run_id=run_id,
            callback_url=callback_url,
            auth=new_auth,
        )

    def handle_event(self, event: Event):
        payload: Optional[Payload] = None
        if isinstance(event, StepReportChangeEvent):
            payload = StepReportPayload(
                step_report=event.step_report,
                run_id=self.run_id,
            )
        elif isinstance(event, NewManifestEvent):
            payload = BotManifestPayload(
                manifest=event.manifest,
                run_id=self.run_id,
            )
        elif isinstance(event, ArtifactsUploadedEvent):
            payload = ArtifactsUploadedPayload(
                run_id=self.run_id,
                output_artifacts_uri=event.output_uri,
            )
        elif isinstance(event, RunStatusChangeEvent):
            payload = RunStatusChangePayload(
                run_id=self.run_id,
                status=event.status,
            )
        else:
            logger.warning(f"Unhandled event: {event}")

        if not payload:
            logger.warning("Could not convert event to a streamable payload")
            return
        self.post(payload=payload)

    def post(self, payload: Payload, **kwargs):
        message_json = payload.__json__()

        try:
            logger.info("Posting streaming message")
            logger.info("URL: %s", self.callback_url)
            logger.info("Payload: %s", message_json)
            response = super().post(
                self.callback_url,
                json=message_json,
                timeout=POST_TIMEOUT_SECONDS,
                **kwargs,
            )
        except Exception:
            # A failed stream message shouldn't break a bot, so catch any possible
            # exception and log it
            logger.exception("Could not post step payload to endpoint")
            return

        logger.info(f"Received response: ({response.status_code}): {response.text}")

        return response


if __name__ == "__main__":
    at = "xxxxxx"
    rt = "yyyyyy"
    _id = "1"
    url = "https://YOUR_URL_ID.execute-api.us-east-1.amazonaws.com/STAGE/webhooks/users-processes-updates/jwt"
    ref_ur = (
        "https://YOUR_URL_ID.execute-api.us-east-1.amazonaws.com/STAGE/refresh-token"
    )

    callback = Streamer.from_encoded_tokens(
        run_id=_id,
        callback_url=url,
        access_token=at,
        refresh_token=rt,
        refresh_url=ref_ur,
    )
    callback.post_status_change(Status.SUCCEEDED)



import os


class EnvironmentVariables:
    """
    Environment variables used by the supervisor package. These are consolidated
    here to have a single source of truth.
    """

    @property
    def run_id(self) -> str:
        """
        Returns:
          (str) The service provider's (Opus or Robocorp) unique ID for this run
        """
        # RC_PROCESS = Robocorp Cloud Process
        # THOUGHTFUL = Opus
        return os.environ.get("RC_PROCESS_RUN_ID") or os.environ.get(
            "THOUGHTFUL_RUN_ID"
        )

    @property
    def callback_url(self) -> str:
        """
        Returns:
          (str) The Fabric endpoint that can receive streaming updates
        """
        return os.environ.get("SUPERVISOR_CALLBACK_URL") or os.environ.get(
            "THOUGHTFUL_CALLBACK_URL"
        )

    @property
    def s3_bucket_uri(self) -> str:
        """
        Returns:
          (str) The location in AWS S3 to which Supervisor should upload
          artifacts to
        """
        return os.environ.get("SUPERVISOR_ARTIFACT_UPLOAD_URI")
